pub mod module;
pub mod op_codes;
pub mod platform;
pub mod types;
pub mod util;
pub mod values;
mod mem;

use crate::vm::mem::DmAllocObject;
use crate::vm::module::DmModule;
use crate::vm::platform::init_platform_functions;
use crate::vm::types::{DmFn, DmImplementation};
use op_codes::*;
use std::alloc::{alloc_zeroed, dealloc, Layout};
use std::collections::HashMap;
use types::DmType;
use util::{get_32_le, get_64_le};
use values::DmValue;
use values::DmValue::*;

pub type PlatformFunction = fn(args: Vec<DmValue>, &mut DmVirtualMachine) -> DmValue;

enum CallFrame {
    PlatformCall(PlatformCallFrame),
    DeimosCall(DeimosCallFrame)
}

pub struct PlatformCallFrame {
    pub name: String,
    pub args: Vec<DmValue>,
}

struct DeimosCallFrame {
    return_address: usize,
}

pub struct DmVirtualMachine<'a> {
    modules: Vec<DmModule<'a>>,
    platform_functions: HashMap<String, PlatformFunction>,
    ip: usize,
    call_stack: Vec<CallFrame>,
    registers: Vec<DmValue>,
    register_state_stack: Vec<Vec<DmValue>>,
}

impl DmVirtualMachine<'_> {
    pub fn new(modules: Vec<DmModule>) -> DmVirtualMachine {
        DmVirtualMachine {
            modules,
            ip: 0,
            registers: Vec::new(),
            platform_functions: init_platform_functions(),
            call_stack: Vec::new(),
            register_state_stack: Vec::new(),
        }
    }

    pub fn call(
        &mut self,
        dm_function: &DmFn,
        args: Vec<DmValue>,
    ) -> DmValue {
        // save current state
        self.call_stack.push(CallFrame::DeimosCall(DeimosCallFrame {
            return_address: self.ip,
        }));
        self.register_state_stack.push(self.registers.clone());

        // zero registers and make sure there are enough for dm_function
        self.registers.clear();
        self.registers.resize(args.len(), DmValue::DmUnit);

        // push args
        for i in 0..args.len() {
            self.registers.insert(i, args[i].clone());
        }

        // run the byte code
        self.run(&dm_function.byte_code);

        // restore state
        self.registers = self.register_state_stack.pop().unwrap();
        if let CallFrame::DeimosCall(deimos_call_frame) = self.call_stack.pop().unwrap() {
            self.ip = deimos_call_frame.return_address;
        }

        // return result
        self.registers.get(0).unwrap().clone()
    }

    pub fn run(&mut self, code: &Vec<u8>) {
        let mut i = 0;
        while i < code.len() {
            match code[i] {
                MOV_INT => {
                    let target_register = code[i + 1] as usize;
                    let operand = get_32_le!(code, i, 2, u32);
                    self.registers.insert(target_register, DmInt(operand as i32));
                    i += 6;
                }
                MOV_LONG => {
                    let target_register = code[i + 1] as usize;
                    let operand = get_64_le!(code, i, 2, u64);
                    self.registers.insert(target_register, DmLong(operand as i64));
                    i += 10;
                }
                MOV_DOUBLE => { /* todo */ }
                MOV_REGISTER => {
                    let target_register = code[i + 1] as usize;
                    let source_register = code[i + 2] as usize;
                    let source_value = self.registers.get(source_register).unwrap();
                    self.registers.insert(target_register, source_value.clone());
                    i += 3;
                }
                ALLOC => {
                    i += 1;
                    let target_register = code[i] as usize;
                    i += 1;
                    let symbol_name_length = get_32_le!(code, i, 0, usize);
                    i += 4;
                    let raw_symbol_name = code[i..(i + symbol_name_length)].to_vec();
                    i += symbol_name_length;
                    let symbol_name = String::from_utf8(raw_symbol_name).unwrap();

                    // find implementation
                    let mut implementation_result: Option<&DmImplementation> = None;
                    'module_find: for module in self.modules.iter() {
                        if let Some(dm_implementation) = module.implementations.get(&symbol_name) {
                            implementation_result = Some(dm_implementation);
                            break 'module_find;
                        }
                    }
                    if implementation_result == None {
                        panic!("Implementation {} not found", symbol_name);
                    }

                    let dm_implementation = implementation_result.unwrap();
                    let size = dm_implementation.size_in_bytes;
                    let layout = Layout::from_size_align(size, 4).unwrap();
                    let pointer = unsafe { alloc_zeroed(layout) };
                    let dm_object = Box::new(DmAllocObject {
                        object_type: dm_implementation,
                        data: pointer,
                        size,
                        layout,
                    });
                    let dm_object_pointer = Box::into_raw(dm_object);
                    self.registers.insert(target_register, DmPointer(dm_object_pointer));
                    i += 6;
                }
                DEALLOC => {
                    let target_register = code[i + 1] as usize;
                    let target_value = self.registers.get(target_register).unwrap();
                    match target_value {
                        DmPointer(ptr) => unsafe {
                            let dm_object = Box::from_raw(ptr.clone());
                            dealloc(dm_object.data, dm_object.layout);
                        }
                        _ => {
                            panic!("Attempt to deallocate at the address stored in a register that is not a pointer.");
                        }
                    }
                    self.registers.insert(target_register, DmUnit);
                    i += 2;
                }
                MOV_INT_TO => {
                    let target_register = code[i + 1] as usize;
                    let target_value = self.registers.get(target_register).unwrap();
                    if let DmPointer(dm_object_ptr) = target_value {
                        let offset = get_32_le!(code, i, 2, isize);
                        let new_address = unsafe {
                            let dm_object = Box::from_raw(dm_object_ptr.clone());
                            let data_ptr = dm_object.data;
                            data_ptr.offset(offset).write(code[i + 6]);
                            data_ptr.offset(offset + 1).write(code[i + 7]);
                            data_ptr.offset(offset + 2).write(code[i + 8]);
                            data_ptr.offset(offset + 3).write(code[i + 9]);
                            Box::into_raw(dm_object)
                        };
                        self.registers.insert(target_register, DmPointer(new_address));
                        i += 10;
                    } else {
                        panic!("target_register {} is not a Pointer", target_register);
                    }
                }
                PLATFORM_CALL => {
                    i += 1;

                    let symbol_name_length = get_32_le!(code, i, 0, usize);
                    i += 4;

                    let symbol_name_raw = code[i..(i + symbol_name_length)].to_vec();
                    i += symbol_name_length;

                    let symbol_name = String::from_utf8(symbol_name_raw).unwrap();

                    let return_register = code[i + 1] as usize;
                    i += 1;

                    let arg_registers_length = code[i] as usize;
                    i += 1;

                    let arg_registers = code[i..(i + arg_registers_length)].to_vec();
                    i += arg_registers_length;

                    let mut args = Vec::new();
                    for arg_register in arg_registers {
                        let register_value = self.registers.get(arg_register as usize).unwrap();
                        args.push(register_value.clone());
                    }

                    let platform_function_result = self.platform_functions.get(&symbol_name);
                    if platform_function_result.is_none() {
                        panic!("Unknown platform function {}", symbol_name);
                    }
                    let platform_function = platform_function_result.unwrap();

                    let call_result = platform_function(args, self);
                    self.registers.insert(return_register, call_result);
                }
                _ => panic!("Invalid code instruction"),
            }
        }
    }

}

#[cfg(test)]
mod dvm_run_tests {
    use super::*;
    use std::alloc::dealloc;

    macro_rules! assert_register {
        ( $expected: expr, $register: expr ) => {
            assert_eq!($expected, $register.unwrap().clone());
        };
    }

    #[test]
    fn mov_1_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 1);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run(&code);
        assert_register!(DmInt(1), vm.registers.get(0));
    }

    #[test]
    fn move_65535_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run(&code);
        assert_register!(DmInt(0xffff), vm.registers.get(0));
    }

    #[test]
    fn move_int_max_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0x0fff_ffff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run(&code);
        assert_register!(DmInt(0x0fff_ffff), vm.registers.get(0));
    }

    #[test]
    fn move_register() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 1, 1);
        add_mov_register(&mut code, 0, 1);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.registers.resize(2, DmUnit);
        vm.run(&code);
        assert_register!(DmInt(1), vm.registers.get(0));
    }

    #[test]
    #[ignore]
    fn mov_int_to_register_as_address() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_mov_int_to(&mut code, 0, 0, 0xff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run(&code);
        let box_address = vm.registers.get(0).unwrap().clone();
        match box_address {
            DmPointer(ptr) => unsafe {
                let dm_object = Box::from_raw(ptr);
                assert_eq!(0xff, *dm_object.data);
                dealloc(dm_object.data, dm_object.layout);
            },
            _ => panic!("Target register is not a pointer"),
        }
    }

    #[test]
    #[ignore]
    fn alloc_and_dealloc_expect_register_cleared() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_dealloc(&mut code, 0);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run(&code);
        assert_register!(DmUnit, vm.registers.get(0));
    }
}
